#! /usr/local/bin/ruby
#
# ndtp.rb -- Network Dictionary Transfer Protocol (NDTP) client
#
# Author: IPR -- Internet Programming with Ruby -- writers
# Copyright (C) 2000-2002 WATANABE Tetsuya, GOTOU Yuuzou
#
# $IPR: ndtp.rb,v 1.12 2003/06/18 16:19:54 gotoyuzo Exp $

=begin

== NDTP

NDTP implements Network Dictionary Transfer Protocol (NDTP) clients.

=== The protocol specification

 * http://www.sra.co.jp/people/m-kasahr/ndtpd/
 * http://www.sra.co.jp/people/m-kasahr/ndtpd/protocol/dserver-protocol2.txt
 * http://www.sra.co.jp/people/m-kasahr/ndtpd/protocol/gaiji-bitmap.txt     
 * http://www.sra.co.jp/people/m-kasahr/ndtpd/protocol/keyword-search.txt

=== Example

  require 'ndtp'
  ndtp = NDTP.new(:Host  => "localhost",
                  :Port  => "ndtp",  # of 2010
                  :User  => "jrn@example.com",
                  :Debug => $DEBUG)
  ndtp.select_dict(1)
  entries = ndtp.entry("test")
  entries.each{|entry, point|
    puts "=> #{entry.gsub(NDTP::GAIJI_REGEX, "")}"
    result = ndtp.consult(point)
    result.gsub!(NDTP::GAIJI_REGEX, "")
    result.gsub!(NDTP::REFERENCE_REGEX, "")
    print(result)
  }
  ndtp.quit

=end

require 'socket'
require 'thread'
require 'nkf'

class NDTP
  GAIJI_REGEX = /<gaiji:([0-9a-z]+)>|\#[0-9a-z]{4}/
  REFERENCE_REGEX = /<([\da-f]+:[\da-f]+)>/
  LF = "\x0A"

  class NdtpError < StandardError; end

  attr_reader :dict_list

  def initialize(config={})
    @config = config
    @mutex = Mutex.new
    start()
  end

  def start
    host = @config[:Host] || "localhost"
    port = @config[:Port] || "ndtp"
    @sock.close if @sock
    @sock = TCPSocket.open(host, port)
    @sock.sync = true

    chk_protocol()

    unless user = @config[:User]
      user = ENV['USER'] || ENV['LOGNAME'] || 'noname'
      user = "#{user}@#{Socket::gethostname()}"
    end
    user_auth(user)

    get_dict_list()
  end

  def chk_protocol()
    if /ndtpd/ =~ version()
      @protocol = :ndtpd
    else
      @protocol = :dserver
    end
  end

  def user_auth(name)
    send('A', name)
    case recv()
    when /^\$A/
      return
    when /^\$N/
      raise NdtpError, 'authentication failure'
    else
      raise NdtpError, 'ndtp protocol error'
    end
  end

  def get_dict_list
    ret = []
    send("t")
    while result = recv()
      break if /^\$\*/ =~ result
      number, name, dict = result.split
      ret << [ number.to_i, dict, name ]
    end
    @dict_list = ret
    ret
  end

  def synchronize
    @mutex.synchronize{
      yield
    }
  end

  def select_dict(dict)
    if dict.is_a?(String)
      if dlist = @dict_list.rassoc(dict)
        dict = dlist[0]
      else
        dict = no.to_i
      end
    end
    send("L#{dict}")
    unless recv() =~ /\$\*/
      raise NdtpError, "cannot select dictionary"
    end
  end

  def entry(word)
    ret = []
    dup_chk = {}

    word = NKF::nkf("-e", word) 
    if @protocol == :ndtpd
      case word 
      when /^\*/    
        word = word.split(//e).reverse.join
        cmdlist = [ 'K', 'A' ]
      else 
        cmdlist = [ 'k', 'a' ]
      end
    else
      cmdlist = [ 'a' ]
    end

    for command in cmdlist
      send('P', command, word)
      count = 0
      index = nil

      while result = recv()
        result.chomp!
        case result
        when /^\$\$/ then break
        when /^\$0/  then next
        when /^\$\?/ then raise NdtpError, "bad argument: #{word}"
        end
        if count % 2 == 0
          index = result
        elsif not dup_chk.key?(result)
          ret << [ index, result ]
          dup_chk[result] = true
        end
        count += 1
      end
    end

    ret
  end

  def consult(point)
    ret = ''
    send('S', point)
    while result = recv()
      break if /^\$\$/ =~ result
      next if /^\$1/ =~ result
      ret << result
    end
    ret
  end

  def get_gaiji_size
    ret = []
    send("XI")
    raise NdtpError, "cannot set gaiji size" unless recv() =~ /^\$I/
    while result = recv()
      break if /^\$\$/ =~ result
      ret << result.chomp
    end
    ret
  end

  def set_gaiji_size(size)
    if size
      send("XL", size)
      unless /^\$\*/ =~ recv()
        raise NdtpError, "cannot set gaiji size"
      end
    end
  end

  def get_gaiji(ids)
    ret = []
    ids.each{|id|
      str = ""
      send("Xb", id)
      unless /^\$I/ =~ recv()
        raise NdtpError, "cannot get gaiji"
      end
      while result = recv()
        break if /^\$\$/ =~ result
        str << result
      end
      ret << str
    }
    ret
  end

  def get_gaiji_all
    unless block_given?
      ret = {}
      get_gaiji_all{|id, value|
        ret[id] = value
      }
      return ret
    end

    ret = {}
    id = value = nil
    send("XB")
    unless /^\$I/ =~ recv()
      raise NdtpError, "cannot get gaiji"
    end
    while result = recv()
      case result
      when /^\$\$/
        break
      when /^\$=(.*)/
        yield(id, value) if id
        id = $1
        value = ''
      else
        value << result
      end
    end
  end

  def quit
    send("Q")
    @dict_list = []
    close
  end

  def version
    send("v")
    recv()
  end

  def close
    @sock.close
    @sock = nil
  end

  private

  def send(*args)
    str = args.join + "\n"
    $stderr.puts "NDTP send: #{str}" if @config[:Debug]
    @sock.print(str)
  end

  def recv()
    str = @sock.gets(LF)
    $stderr.puts "NDTP recv: #{str}" if @config[:Debug]
    str
  end
end

if __FILE__ == $0
  require 'readline'

  NDTPSERVER = ENV['NDTPSERVER'] || "localhost"
  PAGER      = ENV['PAGER'] || "more"
  $KCODE     = 'EUC'

  def pipe_open(cmd)
    rd, wr = IO.pipe
    case pid = fork
    when nil    # child
      wr.close
      STDIN.reopen(rd)
      exec(cmd)
      # NOT REACHED
    else
      begin
        rd.close
        yield(wr)
      ensure
        wr.close
        Process::waitpid(pid,0)
      end
    end
  end

  def print_list(list)
    list.each{|i, name, jname|
      puts "#{i}: #{jname}"
    }
  end

  ndtp = NDTP.new(:Host=>NDTPSERVER, :Debug=>$DEBUG)
  list = ndtp.dict_list
  print_list(list)
  while line = Readline::readline("select> ", true)
    case line
    when /^$/
      print_list(list)
    when /^\d+$/
      dict = line.to_i
      if dict_names = list.assoc(dict)
        ndtp.select_dict(dict)
        name = dict_names[2]
        while line = Readline::readline("#{name}> ", true)
          line.chomp!
          entries = ndtp.entry(line)
          next if entries.empty?
          pipe_open(PAGER){|io|
            entries.each{|index, point|
              buf = ndtp.consult(point)
              buf.gsub!(NDTP::GAIJI_REGEX, "")
              buf.gsub!(NDTP::REFERENCE_REGEX, "")
              io << buf << "==\n"
            }
          }
        end
        puts ""
        print_list(list)
      end
    end
  end
  ndtp.quit
end
